from .distn import RegressionDistn, ClassificationDistn
from .normal import Normal, NormalFixedVar
from .multivariate_normal import MultivariateNormal
from .lognormal import LogNormal
from .exponential import Exponential
from .categorical import k_categorical, Bernoulli
